@@ -1,4 +1,4 @@
-static char *rcsid="$Id: stdutil.c,v 1.8 1996/07/30 02:25:43 nneul Exp $";
+static char *rcsid="$Id: stdutil.c,v 1.11 1997/11/16 21:54:53 nneul Exp $";
 
 /**
  **  File: stdutil.c
@@ -18,7 +18,7 @@ char *strdup(char *str)
 	temp = (char *) malloc ( strlen(str) + 1 );
 	if (!temp)
 	{
-		DoPError("Couldn't malloc memory for string!");
+		MSG_Error_SystemError("Couldn't malloc memory for string!");
 	}
 	strcpy(temp,str);
 
