@@ -1,4 +1,4 @@
-static char *rcsid="$Id: util.c,v 1.14 1996/08/18 00:41:24 nneul Exp $";
+static char *rcsid="$Id: util.c,v 1.38 1998/10/07 23:58:36 nneul Exp $";
 
 /**
  **  File: util.c
@@ -7,6 +7,24 @@ static char *rcsid="$Id: util.c,v 1.14 1996/08/18 00:41:24 nneul Exp $";
 
 #include "cgiwrap.h"	/* Headers for all CGIwrap source files */
 
+/*
+ * Safe malloc
+ */
+char *SafeMalloc(size_t size, char *what)
+{
+	char *tmp;
+	
+	tmp = (char *) malloc( size );
+	if ( !tmp )
+	{
+		char msg[500];
+		sprintf(msg, "Couldn't malloc() (%d) bytes for (%s).\n",
+			size, what);
+		MSG_Error_SystemError(msg);
+	}
+	return tmp;
+}
+
 /*
  * Build the ARGV array for passing to the called script
  */
@@ -15,12 +33,7 @@ char **CreateARGV( char *scrStr, int argc, char *argv[])
 	char **temp;
 	int i;
 
-	temp = (char **) malloc( (argc+1) * sizeof(char *) );
-
-	if ( !temp )
-	{
-		DoPError("Couldn't malloc() memory for argv array\n");
-	}
+	temp = (char **) SafeMalloc( (argc+1) * sizeof(char *), "ARGV array");
 	
 	temp[0] = StripPathComponents( CountSubDirs(scrStr), scrStr );
 	temp[argc] = 0;
@@ -52,7 +65,6 @@ char *StripPrefix( char *stringToStrip, char *stripFrom )
 	return stripFrom;
 }
 
-
 /*
  *   Extract and return the value portion of a key=value pair found in a string
  */
@@ -72,7 +84,8 @@ char *GetValue( char *keyword, char *string )
 		dataStr = (char *) strstr(tempStr, keyword);
 		if ( dataStr == tempStr )
 		{
-			returnStr = (char *) strchr (dataStr, '=') + 1;
+			returnStr = (char *) strchr (dataStr, '=');
+			if ( returnStr) returnStr++;
 			tempStr = (char *) 0;
 		}
 		else
@@ -81,7 +94,10 @@ char *GetValue( char *keyword, char *string )
 		}
 	}
 
-	returnStr = strdup(returnStr);
+	if ( returnStr )
+	{
+		returnStr = strdup(returnStr);
+	}
 	free(theString);
 	return returnStr;
 }
@@ -90,23 +106,46 @@ char *GetValue( char *keyword, char *string )
 /*
  * Check if a path is safe to use
  *   Return true if 'path' contains any whitespace or non-printables
- *   Return true if 'path' contains '../' or './'
+ *   Return true if 'path' contains '../'
  */
 int CheckPath(char *path)
 {
-	int i;
+	int i,c;
 
 	for (i=0; i< strlen(path); i++)
 	{
-		if ( !isprint(path[i]) || isspace(path[i]) )
+		c = path[i];
+		if ( !isprint(c) || isspace(c) )
 		{
 			return 1;
 		}
 	}
 
-	return ( strstr(path, "../") || strstr(path, "./") );
+	return ( strstr(path, "../") != NULL );
 }
 
+/*
+ * Condense slashes, removing duplicates and trailers
+ */
+char *CondenseSlashes(char *path)
+{
+	char *tmp;
+	int i,j;
+
+	tmp = (char *) SafeMalloc( strlen(path)+1, "CondenseSlashes");
+
+	for (i=0, j=0; i<strlen(path); i++)
+	{
+		if ( !((path[i] == '/') & 
+			((path[i+1] == '/') | (path[i+1] == '\0')) ))
+		{
+			tmp[j++] = path[i];
+		}
+	}
+	tmp[j] = 0;
+
+	return tmp;
+}
 
 /*
  * Count the number of /'s in a string
@@ -118,7 +157,10 @@ int CountSubDirs(char *path)
 
 	for (i=0; i<strlen(path); i++)
 	{
-		count += (path[i] == '/');
+		if ( path[i] == '/' )
+		{
+			count++;
+		}
 	}
 	return count;
 }
@@ -156,75 +198,102 @@ void ChangeToCGIDir(char *scriptPath)
 	i = CountSubDirs(scriptPath) - 1;
 	tempdir = GetPathComponents(i, scriptPath);
 
-	tempstring = (char *) malloc(strlen(tempdir) + 5);
+	tempstring = (char *) SafeMalloc (strlen(tempdir) + 5,
+		"cgi directory path");
 	tempstring[0] = '/';
 	tempstring[1] = 0;
 	strcat(tempstring, tempdir);
 
-	DEBUG_Str("\nChanging Current Directory to", tempstring);
+	DEBUG_Str("\nChanging current directory to", tempstring);
 	chdir(tempstring);
 
 	free(tempdir);
 	free(tempstring);
 }
 
-
 /*
  * Perform checks on the userid 
  */
 void CheckUser(struct passwd *user)
 {
-	int deny_exists, allow_exists;
+#if defined(CONF_DENYFILE) || defined(CONF_ALLOWFILE)
+	int deny_exists=0, allow_exists=0;
 	int in_deny, in_allow;	
 
-#if defined(CONF_ACCESS)
 #if defined(CONF_CHECKHOST)
 	if ( ( !getenv("REMOTE_ADDR") ) | (getenv("REMOTE_ADDR")[0] == '\0') )
 	{
 		Log(user->pw_name, "-", "no remote host");
-		DoError("Your host (null) is not allowed to run this");
+		MSG_Error_General("Your host (null) is not allowed to run this");
 	}
 #endif
 
+#if defined(CONF_DENYFILE)
 	deny_exists = FileExists(CONF_DENYFILE);
+#endif
+#if defined(CONF_ALLOWFILE)
 	allow_exists = FileExists(CONF_ALLOWFILE);
+#endif
 	in_deny = 0;
 	in_allow = 0;
 
+#if defined(CONF_DENYFILE)
 	if ( deny_exists )
 	{
 		in_deny = UserInFile(CONF_DENYFILE, user->pw_name);
 	}
-
+#endif
+#if defined(CONF_ALLOWFILE)
 	if ( allow_exists )
 	{
 		in_allow = UserInFile(CONF_ALLOWFILE, user->pw_name);
 	}
+#endif
 
 	if ( !deny_exists && !allow_exists )
 	{
 		Log(user->pw_name, "-", "access control files not found");
-		DoError("Access control files not found!");
+		MSG_Error_AccessControl("Access control files not found!");
 	}
 
-	if ( in_allow && in_deny )
+	if ( (in_allow && in_deny) ||
+		( allow_exists && !in_allow ) ||
+		( deny_exists && in_deny ) )
 	{
-		Log(user->pw_name, "-", "user in both allow and deny files");
-		DoError("User is both allowed and denied!");
-	}
-
-	if ( allow_exists && !in_allow )
-	{
-		Log(user->pw_name, "-", "user not in allow file");
-		DoError("User not in allow file!");
-	}
-
-	if ( deny_exists && in_deny )
-	{
-		Log(user->pw_name, "-", "user in deny file");
-		DoError("User in deny file!");
+		Log(user->pw_name, "-", "user/host not permitted");
+		MSG_Error_AccessControl("Script userid and/or remote host not permitted!");
 	}
 #endif
+
+#if defined(CONF_MINIMUM_UID)
+	if ( user->pw_uid < CONF_MINIMUM_UID )
+	{
+		Log(user->pw_name, "-", "uid less than minimum");
+		MSG_Error_AccessControl("UID of script userid less than configured minimum.");
+	}
+#endif
+
+#if defined(CONF_CHECKSHELL)
+	{
+	char *sh, *getusershell();
+	int found = 0;
+
+	while ( sh = getusershell() )
+	{
+		if (0 == strcmp( sh, user->pw_shell ))
+		{
+		 	found = 1;
+			break;
+		}
+	}
+	if (found == 0)
+	{
+	  Log(user->pw_name, "-", "restricted login shell");
+	  MSG_Error_AccessControl("Restricted login shell, permission denied.");
+	}
+	}
+#endif
+
 }
 
 
@@ -232,55 +301,75 @@ void CheckUser(struct passwd *user)
 /*
  * Perform file checks on the script file
  */
-void CheckScriptFile(struct passwd *user, char *scriptPath)
+void CheckScriptFile(void)
 {
 	struct stat fileStat; /* For checking file status */
-	char tempErrString[255];
+	struct stat fileLStat; /* For checking symlink status */
 
-	if ( CheckPath(scriptPath) )
+	if ( CheckPath(Context.scriptFullPath) )
 	{
-		DoError("Script path contains illegal characters or '.' or '..'!");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script path contains illegal components");
 	}
 	
 #if !defined(CONF_SUBDIRS)
 	/* Check if script is in a subdirectory */
-	if ( CountSubDirs(scriptPath) > 0 )
+	if ( CountSubDirs(Context.scriptRelativePath) > 0 )
 	{
-		Log(user->pw_name, scriptPath, "script in subdir not allowed");
-		DoError("Script is in sub-directory, that is not allowed.");
+		Log(Context.user.pw_name, Context.scriptFullPath, 
+			"script in subdir not allowed");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Scripts in subdirectories are not allowed");
 	}
 #endif
 
-	if ( stat(scriptPath, &fileStat) )
+	if ( stat(Context.scriptFullPath, &fileStat) )
 	{
-		DoPError("Script file not found!");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script file not found.");
 	}
 
+#if defined(CONF_CHECK_SYMLINK)
+	if ( lstat(Context.scriptFullPath, &fileLStat) )
+	{
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script file not found.");
+	}
+
+	if ( S_ISLNK(fileLStat.st_mode) )
+	{
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is a symbolic link");
+	}
+#endif		
+	
 	if ( !S_ISREG(fileStat.st_mode) )
 	{
-		DoError("Script is not a regular file");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is not a regular file");
 	}
 
 	if (!(fileStat.st_mode & S_IXUSR))
 	{
-		sprintf(tempErrString, "Script is not executable. Issue chmod 755 %s", 
-			scriptPath);
-		DoError(tempErrString);
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is not executable. Issue 'chmod 755 filename'");
 	}
 
 
 #if defined(CONF_CHECK_SCRUID)
-	if (fileStat.st_uid != user->pw_uid)
+	if (fileStat.st_uid != Context.user.pw_uid)
 	{
-		DoError("Script does not have same UID");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script does not have same UID");
 	}
 #endif
 
 
 #if defined(CONF_CHECK_SCRGID)
-	if (fileStat.st_gid != user->pw_gid)
+	if (fileStat.st_gid != Context.user.pw_gid)
 	{
-		DoError("Script does not have same GID");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script does not have same GID");
 	}
 #endif
 
@@ -288,7 +377,8 @@ void CheckScriptFile(struct passwd *user, char *scriptPath)
 #if defined(CONF_CHECK_SCRSUID)
 	if (fileStat.st_mode & S_ISUID)
 	{
-		DoError("Script is SUID - Will not Execute!");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is setuid");
 	}
 #endif
 
@@ -296,10 +386,26 @@ void CheckScriptFile(struct passwd *user, char *scriptPath)
 #if defined(CONF_CHECK_SCRSGID)
 	if (fileStat.st_mode & S_ISGID)
 	{
-		DoError("Script is SGID - Will not Execute!");
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is setgid");
 	}
 #endif
 
+#if defined(CONF_CHECK_SCRGWRITE)
+	if (fileStat.st_mode & S_IWGRP)
+	{
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is group writable.");
+	}
+#endif
+
+#if defined(CONF_CHECK_SCROWRITE)
+	if (fileStat.st_mode & S_IWOTH)
+	{
+		MSG_Error_ExecutionNotPermitted(Context.scriptFullPath,
+			"Script is world writable.");
+	}
+#endif
 }
 
 
@@ -308,17 +414,17 @@ void CheckScriptFile(struct passwd *user, char *scriptPath)
  */
 void VerifyExecutingUser(void)
 {
+#if defined(CONF_CHECK_HTTPD_USER)
 	struct passwd *user;
 
-#if defined(CONF_CHECK_HTTPD_USER)
-        if ( !(user = getpwnam(CONF_HTTPD_USER)) )
-        {
-                DoError("Configured server userid not found.");
-        }
+	if ( !(user = getpwnam(CONF_HTTPD_USER)) )
+	{
+		MSG_Error_ServerUserNotFound();
+	}
 
 	if ( getuid() != user->pw_uid )
 	{
-		DoError("Real UID does not match configured server userid!.");
+		MSG_Error_ServerUserMismatch();
 	}
 #endif
 }
@@ -330,15 +436,10 @@ void VerifyExecutingUser(void)
 char *BuildScriptPath(char *basedir, char *scrStr)
 {
 	char *tmp;
-	tmp = (char *) malloc( strlen(basedir) + strlen(scrStr) + 5);
-
-	if ( !tmp )
-	{
-		DoPError("Couldn't malloc memory for scriptPath");
-	}
+	tmp = (char *) SafeMalloc ( strlen(basedir) + strlen(scrStr) + 5,
+		"scriptPath");
 
 	sprintf(tmp, "%s/%s", basedir, scrStr);
-
 	return tmp;
 }
 
@@ -440,29 +541,109 @@ char *StripPathComponents(int count, char *path)
 }
 
 
+/*
+ * Set Environment Variables
+ */
+void SetEnvironmentVariables(void)
+{
+#if defined(CONF_SETENV_ANY)
+	int i;
+	char msg[200];
+
+	struct cgiwrap_setenv_table
+	{
+		char *variable;
+		char *value;
+		char *setstring;
+	} cgiwrap_setenvs[] = {
+#if defined(CONF_SETENV_PATH)
+		{"PATH", CONF_SETENV_PATH, "PATH=" CONF_SETENV_PATH},
+#endif
+#if defined(CONF_SETENV_TZ)
+		{"TZ", CONF_SETENV_TZ, "TZ=" CONF_SETENV_TZ},
+#endif
+		{0,0,0}
+	};
+
+	for (i=0; cgiwrap_setenvs[i].variable; i++)
+	{
+		sprintf(msg, "Setting Environment Variable(%s) to (%s)\n", 
+			cgiwrap_setenvs[i].variable,
+			cgiwrap_setenvs[i].value);
+		DEBUG_Msg(msg);
+
+#if defined(HAS_PUTENV)
+		putenv(cgiwrap_setenvs[i].setstring);
+#endif
+	}
+#endif
+}
+
+
 /*
  * Set Resource Limits
  */
 void SetResourceLimits(void)
 {
+#if defined(CONF_USE_RLIMIT_ANY)
+	struct rlimit limstruct;
+	int i;
+	char msg[200];
+
+	struct cgiwrap_rlimit_table
+	{
+		int which;
+		char *label;
+		long value;
+	} cgiwrap_rlimits[] = {
 #if defined(CONF_USE_RLIMIT_CPU) && defined(RLIMIT_CPU)
-	/* CPU Time Limit */
-	struct rlimit cpulim = { 9, 10 	}; 
+		{RLIMIT_CPU, "cpu time", CONF_USE_RLIMIT_CPU},
 #endif
-
 #if defined(CONF_USE_RLIMIT_VMEM) && defined(RLIMIT_VMEM)
-	/* Virtual Memory Limit */
-	struct rlimit vmemlim = { 2000000, 2500000 };
+		{RLIMIT_VMEM, "virtual memory", CONF_USE_RLIMIT_VMEM},
 #endif
-
-#if defined(CONF_USE_RLIMIT_CPU) && defined(RLIMIT_CPU)
-	DEBUG_Msg("\nSetting Limits (CPU Usage)\n");
-	setrlimit( RLIMIT_CPU, &cpulim );
+#if defined(CONF_USE_RLIMIT_FSIZE) && defined(RLIMIT_FSIZE)
+		{RLIMIT_FSIZE, "writable file size", CONF_USE_RLIMIT_FSIZE},
 #endif
+#if defined(CONF_USE_RLIMIT_DATA) && defined(RLIMIT_DATA)
+		{RLIMIT_DATA, "data size", CONF_USE_RLIMIT_DATA},
+#endif
+#if defined(CONF_USE_RLIMIT_STACK) && defined(RLIMIT_STACK)
+		{RLIMIT_STACK, "stack size", CONF_USE_RLIMIT_STACK},
+#endif
+#if defined(CONF_USE_RLIMIT_AS) && defined(RLIMIT_AS)
+		{RLIMIT_AS, "total available memory", CONF_USE_RLIMIT_AS},
+#endif
+#if defined(CONF_USE_RLIMIT_CORE) && defined(RLIMIT_CORE)
+		{RLIMIT_CORE, "core file size", CONF_USE_RLIMIT_CORE},
+#endif
+#if defined(CONF_USE_RLIMIT_RSS) && defined(RLIMIT_RSS)
+		{RLIMIT_RSS, "resident set size", CONF_USE_RLIMIT_RSS},
+#endif
+#if defined(CONF_USE_RLIMIT_NPROC) && defined(RLIMIT_NPROC)
+		{RLIMIT_NPROC, "number of processes", CONF_USE_RLIMIT_NPROC},
+#endif
+#if defined(CONF_USE_RLIMIT_NOFILE) && defined(RLIMIT_NOFILE)
+		{RLIMIT_NOFILE, "number of open files", CONF_USE_RLIMIT_NOFILE},
+#endif
+#if defined(CONF_USE_RLIMIT_MEMLOCK) && defined(RLIMIT_MEMLOCK)
+		{RLIMIT_MEMLOCK, "lockable memory", CONF_USE_RLIMIT_MEMLOCK},
+#endif
+		{0,0,0}
+	};
 
-#if defined(CONF_USE_RLIMIT_VMEM) && defined(RLIMIT_VMEM)
-	DEBUG_Msg("Setting Limits (Virtual Memory)\n");
-	setrlimit( RLIMIT_VMEM, &vmemlim );
+	for (i=0; cgiwrap_rlimits[i].label; i++)
+	{
+		limstruct.rlim_cur = cgiwrap_rlimits[i].value;
+		limstruct.rlim_max = cgiwrap_rlimits[i].value;
+
+		sprintf(msg, "Limiting (%s) to (%ld)\n", 
+			cgiwrap_rlimits[i].label,
+			cgiwrap_rlimits[i].value);
+		DEBUG_Msg(msg);
+
+		setrlimit(cgiwrap_rlimits[i].which, &limstruct);
+	}
 #endif
 }
 
@@ -501,7 +682,7 @@ void ChangeID ( struct passwd *user)
 	seteuid( user->pw_uid );
 	setruid( user->pw_uid );
 #else
-	DoError("Configuration Error, No Way to Change IDs");
+#error "Configuration Error, No Way to Change IDs"
 #endif
 
 
@@ -517,19 +698,19 @@ void ChangeID ( struct passwd *user)
 	/***/
 	if ( getuid() != user->pw_uid )
 	{
-		DoError("Real UID could not be changed!");
+		MSG_Error_General("Real UID could not be changed!");
 	}
 	if ( geteuid() != user->pw_uid )
 	{
-		DoError("Effective UID could not be changed!");
+		MSG_Error_General("Effective UID could not be changed!");
 	}
 	if ( getgid() != user->pw_gid )
 	{
-		DoError("Real GID could not be changed!");
+		MSG_Error_General("Real GID could not be changed!");
 	}
 	if ( getegid() != user->pw_gid )
 	{
-		DoError("Effective GID could not be changed!");
+		MSG_Error_General("Effective GID could not be changed!");
 	}
 }
 
@@ -540,12 +721,12 @@ void ChangeAuxGroups(struct passwd *user)
 {
 #if defined(HAS_SETGROUPS) && defined(CONF_SETGROUPS)
 	if ( setgroups(0, NULL) == -1 )
-		DoPError("setgroups() failed!");
+		MSG_Error_SystemError("setgroups() failed!");
 #endif
 
 #if defined(HAS_INITGROUPS) && defined(CONF_INITGROUPS)
 	if ( initgroups( user->pw_name, user->pw_gid ) == -1 )
-		DoPError("initgroups() failed!");
+		MSG_Error_SystemError("initgroups() failed!");
 #endif
 }
 
@@ -556,57 +737,79 @@ void ChangeAuxGroups(struct passwd *user)
 int UserInFile(char *filename, char *user)
 {
 	FILE *file;
-	char temp[200];
+	char temp[HUGE_STRING_LEN], *tmpuser;
 #if defined(CONF_CHECKHOST)
-      int unlen,addr[4],mask[4],snet[4];
-      char *i;
+	int unlen,remote_addr[4],spec_mask[4],spec_addr[4];
+	char *i;
 #endif
+	int j;
 
 #if defined(CONF_CHECKHOST)
-      unlen = strlen(user);
-      mask[0]=255; mask[1]=255; mask[2]=255; mask[3]=255;
-      if (sscanf(getenv("REMOTE_ADDR"),"%d.%d.%d.%d",
-              &addr[0],&addr[1],&addr[2],&addr[3]) != 4 )
-      {
-              Log(user, "-", "no remote host");
-              DoError("Your host (undetermined) is not allowed to run this");
-      }
+	spec_mask[0]=255; 
+	spec_mask[1]=255; 
+	spec_mask[2]=255; 
+	spec_mask[3]=255;
+	if (sscanf(getenv("REMOTE_ADDR"),"%d.%d.%d.%d",
+		&remote_addr[0],&remote_addr[1],
+		&remote_addr[2],&remote_addr[3]) != 4 )
+	{
+		Log(user, "-", "no remote host");
+		MSG_Error_General("Your host is not allowed to run this");
+	}
 #endif
 
 	if ( (file=fopen(filename,"r")) == NULL )
 	{
-		DoPError("Couldn't Open User List File");
+		MSG_Error_SystemError("Couldn't open access control file");
 	}
 
 	while ( !feof(file) )
 	{
-		fscanf(file,"%s",temp);
+		temp[0] = 0;
+		if ( !fgets(temp, HUGE_STRING_LEN, file) )
+		{
+			fclose(file);
+			return 0;
+		}
+		for (j=0; j<=strlen(temp); j++)
+		{
+			if ( !isprint(temp[j]) )
+			{
+				temp[j] = 0;
+			}
+		}
+
 		if ( !strcmp(temp,user) )
 		{
 			fclose(file);
 			return 1;
 		}
 #if defined(CONF_CHECKHOST)
-              else if (( !strncmp(temp,user,unlen) ) && (temp[unlen] == '@') )
-              {
-                      i = &temp[unlen];
-                      while (i != NULL)
-                      {
-                              i++;
-                              sscanf(i,"%d.%d.%d.%d/%d.%d.%d.%d",
-                              &snet[0],&snet[1],&snet[2],&snet[3],
-                              &mask[0],&mask[1],&mask[2],&mask[3]);
+		i = strchr(temp, '@');
+		tmpuser = strtok(temp, "@");
+		if ( (i != NULL) && (tmpuser != NULL) &&
+			( (!strcmp(tmpuser,user)) || (!strcmp(tmpuser, "*")) ) )
+		{
+			while (i != NULL)
+			{
+				i++;
+				sscanf(i,"%d.%d.%d.%d/%d.%d.%d.%d",
+					&spec_addr[0], &spec_addr[1],
+					&spec_addr[2], &spec_addr[3],
+					&spec_mask[0], &spec_mask[1],
+					&spec_mask[2], &spec_mask[3]);
 
-                              if ((snet[0] == (mask[0] & addr[0])) &&
-                                  (snet[1] == (mask[1] & addr[1])) &&
-                                  (snet[2] == (mask[2] & addr[2])) &&
-                                  (snet[3] == (mask[3] & addr[3])))
-                              {
-                                      return 1;
-                              }
-                      i = strchr(i,',');
-                      }
-              }
+				if ((spec_addr[0] == (spec_mask[0] & remote_addr[0])) &&
+					(spec_addr[1] == (spec_mask[1] & remote_addr[1])) &&
+					(spec_addr[2] == (spec_mask[2] & remote_addr[2])) &&
+					(spec_addr[3] == (spec_mask[3] & remote_addr[3])))
+				{
+					fclose(file);
+					return 1;
+            	}
+    			i = strchr(i,',');
+    		}
+		}
 #endif
 	}
 	fclose(file);
@@ -616,62 +819,47 @@ int UserInFile(char *filename, char *user)
 
 
 /*
- * Send a Content-Type header if one hasn't already be sent
+ * Initialize the logging
  */
-void SendHeader(char *type)
+void LogInit (void)
 {
-	static int done = 0;
-
-	if ( !done )
-	{
-		printf ("Content-Type: %s\n", type);
-		done = 1;
-	}
-}
-
-
-/*
- * Output an error message with system error message string
- */
-void DoPError (char *msg)
-{
-	SendHeader("text/plain");
-	printf("\n");
-
-	printf("CGIwrap Error: %s\n", msg);
-
-#if defined(HAS_STRERROR)
-	printf("Error: %s (%d)\n", strerror(errno), errno);
-#elif defined(HAS_PERROR)
-	perror("Error");
-#else
-	printf("Error: %d\n", errno);
+#if defined(CONF_LOG_USEFILE)
+	FILE *logFile;
+	int logfd;
 #endif
 
-	exit(1);
+	DEBUG_Msg("Initializing Logging");
+
+#if defined(CONF_LOG_USEFILE)
+	logfd = open(CONF_LOG_LOGFILE, O_WRONLY | O_CREAT | O_APPEND, 
+		S_IRUSR | S_IWUSR);
+	if ( logfd == -1 )
+	{
+		MSG_Error_SystemError("Could not open log file for appending!");
+	}
+	/* set the close-on-exec flag for logfd */
+	fcntl(logfd, F_SETFD, 1);
+
+	/* open a file pointer from that file descriptor */
+	logFile = fdopen(logfd, "a");
+	if ( !logFile )
+	{
+		MSG_Error_SystemError("Could not open file stream from file descriptor!");
+	}
+
+	Context.logFile = logFile;
+#endif
+#if defined(CONF_LOG_USESYSLOG) && defined(HAS_SYSLOG)
+	openlog("cgiwrap", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
+#endif
 }
 
-
-/*
- * Output a error message
- */
-void DoError (char *msg)
-{
-	SendHeader("text/plain");
-	printf("\n");
-
-	printf("CGIwrap Error: %s \n", msg);
-	exit(1);
-}
-
-
 /*
  * Add an entry to the log file
  */
 void Log (char *user, char *script, char *msg)
 {
 	time_t timevar;
-	FILE *logFile;
 	char *timeString;
 
 	time(&timevar);
@@ -682,12 +870,7 @@ void Log (char *user, char *script, char *msg)
 #if defined(CONF_LOG_USEFILE)
 	DEBUG_Msg("Logging Request (File)");
 
-	if ( (logFile = fopen(CONF_LOG_LOGFILE, "a")) == NULL )
-	{
-		DoPError("Could not open log file for appending!");
-	}
-	
-	fprintf(logFile, "%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
+	fprintf(Context.logFile, "%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
 		NullCheck( user ),
 		NullCheck( script ),
 		NullCheck( getenv("REMOTE_HOST") ),
@@ -695,14 +878,11 @@ void Log (char *user, char *script, char *msg)
 		NullCheck( getenv("REMOTE_USER") ),
 		NullCheck( msg ),
 		NullCheck( timeString ) );
-
-	fclose(logFile);
+	fflush(Context.logFile);
 #endif
 #if defined(CONF_LOG_USESYSLOG) && defined(HAS_SYSLOG)
 	DEBUG_Msg("Logging Request (syslog)");
 
-	openlog("cgiwrap", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
-
 	syslog(LOG_INFO, "[%s] %s, %s, %s, %s, %s, %s",
 		NullCheck( CONF_LOG_LABEL ),
 		NullCheck( user ), 
@@ -712,10 +892,23 @@ void Log (char *user, char *script, char *msg)
 		NullCheck( getenv("REMOTE_USER") ),
 		NullCheck( msg ) );
 
-	closelog();
+/* don't want to close log at this point */
+/*	closelog(); */
 #endif
 }
 
+/*
+ * Terminate logging
+ */
+void LogEnd(void)
+{
+#if defined(CONF_LOG_USEFILE)
+	if ( Context.logFile )
+	{
+		fclose(Context.logFile);
+	}
+#endif
+}
 
 /*
  * Set the correct SCRIPT_NAME environment variable
@@ -724,14 +917,9 @@ void SetScriptName(char *userStr, char *scrStr )
 {
 	char *buf;
 
-	buf = (char*) malloc (strlen("SCRIPT_NAME") +
+	buf = (char*) SafeMalloc (strlen("SCRIPT_NAME") +
 	    strlen(getenv("SCRIPT_NAME")) + strlen(userStr)
-	    + strlen(scrStr) + 5);
-
-	if( !buf )
-	{
-		DoPError("Couldn't malloc memory for SCRIPT_NAME buf!");
-	}
+	    + strlen(scrStr) + 5, "new SCRIPT_NAME environment variable");
 
 	sprintf(buf, "%s=%s/%s/%s", "SCRIPT_NAME", 
 	    getenv("SCRIPT_NAME"), userStr, scrStr);
@@ -746,13 +934,8 @@ void SetPathTranslated( char *scriptPath )
 {
 #ifndef CONF_FIXED_PATHTRANS
 	char *buf;
-	buf = (char *) malloc( strlen("PATH_TRANSLATED") +
-		strlen(scriptPath) + 5 );
-
-	if( !buf )
-	{
-		DoPError("Couldn't malloc memory for PATH_TRANSLATED buf!");
-	}
+	buf = (char *) SafeMalloc( strlen("PATH_TRANSLATED") +
+		strlen(scriptPath) + 5, "new PATH_TRANSLATED environment variable");
 
 	sprintf(buf, "%s=%s", "PATH_TRANSLATED", scriptPath); 
 	putenv(buf);
@@ -769,13 +952,9 @@ void SetPathTranslated( char *scriptPath )
 		return;
 	}
 
-	buf = (char *) malloc( strlen("PATH_TRANSLATED") +
-		strlen(docroot) + strlen(pathinfo) + 5);
-
-	if( !buf )
-	{
-		DoPError("Couldn't malloc memory for PATH_TRANSLATED buf!");
-	}
+	buf = (char *) SafeMalloc( strlen("PATH_TRANSLATED") +
+		strlen(docroot) + strlen(pathinfo) + 5, 
+		"new PATH_TRANSLATED environment variable");
 
 	sprintf(buf, "%s=%s%s", "PATH_TRANSLATED", docroot, pathinfo); 
 	putenv(buf);
@@ -791,6 +970,9 @@ void SetPathTranslated( char *scriptPath )
  * might make. It also prevents the script from inheriting the servers
  * authentication if it is running authenticated
  */
+#if defined(CONF_AFS_SETPAG)
+	void setpag(void);
+#endif
 void Create_AFS_PAG(void)
 {
 #if defined(CONF_AFS_SETPAG)
@@ -805,16 +987,16 @@ void Create_AFS_PAG(void)
  */
 char *GetUserDir(char *user)
 {
+#if defined(CONF_USERDIRFILE)
 	FILE *file;
 	static char temp[500];
 	int i, j;
 
-#if defined(CONF_USERDIRFILE)
-	DEBUG_Msg("\nProcessing User Directory Configuration File.");
+	DEBUG_Msg("\nProcessing user directory configuration file.");
 
 	if ( (file=fopen(CONF_USERDIRFILE,"r")) == NULL )
 	{
-		DoPError("Couldn't Open User Directory Config File");
+		MSG_Error_SystemError("Couldn't open user directory config file");
 	}
 
 	temp[0]=0;
@@ -849,33 +1031,21 @@ char *GetUserDir(char *user)
  */
 char *GetBaseDirectory(struct passwd *user)
 {
-	char *tmp;
-	char *userdir;
-	char *basedir;
+	char *userdir, *basedir;
 
 	userdir = GetUserDir(user->pw_name);
 
 	if ( userdir )
 	{
 		DEBUG_Msg("Using configured base directory.\n");
-		basedir = (char *) malloc( strlen(userdir) + 4 );
+		basedir = (char *) SafeMalloc ( strlen(userdir) + 4,
+			"script base directory");
 		strcpy(basedir, userdir);
-
-		if ( !basedir )
-		{
-			DoPError("Couldn't malloc memory for basedir");
-		}
 	}
 	else
 	{
-		basedir = (char *) malloc( strlen(user->pw_dir) +
-			strlen(CONF_CGIDIR) + 4 );
-		
-		if ( !basedir )
-		{
-			DoPError("Couldn't malloc memory for basedir");
-		}
-
+		basedir = (char *) SafeMalloc (strlen(user->pw_dir) +
+			strlen(CONF_CGIDIR) + 4, "script base directory");
 		sprintf(basedir, "%s/%s", user->pw_dir, CONF_CGIDIR);
 	}
 
@@ -925,5 +1095,3 @@ int DirExists(char *path)
 
 	return 1;
 }
-
-
