@@ -1,15 +1,17 @@
-/* rcsid: $Id: util.h,v 1.9 1996/05/31 19:19:13 nneul Exp $ */
+/* rcsid: $Id: util.h,v 1.15 1998/01/13 01:48:02 nneul Exp $ */
 
 /**
  **  File: util.h
  **  Purpose: Prototypes and defines for routines in util.c
  **/ 
 
+char *SafeMalloc( size_t size, char *what );
 char **CreateARGV( char *scrStr, int argc, char *argv[]);
 char *StripPrefix(char *, char *);
 void ChangeID(struct passwd *user);
 void ChangeAuxGroups(struct passwd *user);
 void SetResourceLimits(void);
+void SetEnvironmentVariables(void);
 void SetSignals(void);
 char *GetValue(char *keyword, char *string);
 int CountSubDirs(char *path);
@@ -17,7 +19,8 @@ int CheckPath(char *path);
 void ChangeToCgiDir(struct passwd *user);
 void OutputEnvironment(void);
 void CheckUser(struct passwd *user);
-void CheckScriptFile(struct passwd *user, char *scriptPath);
+char *CondenseSlashes(char *path);
+void CheckScriptFile(void); 
 void VerifyExecutingUser(void);
 char *BuildScriptPath(char *basedir, char *scrStr);
 char *GetPathComponents(int count, char *path);
@@ -25,8 +28,8 @@ char *StripPathComponents(int count, char *path);
 void ChangeID ( struct passwd *user);
 void ChangeAuxGroups(struct passwd *user);
 int UserInFile(char *filename, char *user);
-void SendHeader(char *type);
 void DoPError (char *msg);
+void ChangeToCGIDir(char *scriptPath);
 void DoError (char *msg);
 void Log (char *user, char *script, char *msg);
 void SetScriptName(char *userStr, char *scrStr );
@@ -37,3 +40,5 @@ char *GetBaseDirectory(struct passwd *user);
 
 int FileExists(char *path);
 int DirExists(char *path);
+void LogInit (void);
+void LogEnd (void);
